# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""handler.py provides decorators for POST and GET handlers."""

import datetime
import functools
import json
import re

from flask import g
from flask import make_response
from flask import request
import google.auth
from google.auth.transport import requests as google_requests
from google.oauth2 import id_token
import requests

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.system import environment
from libs import access
from libs import auth
from libs import csp
from libs import helpers

JSON = 'json'
FORM = 'form'
HTML = 'html'
TEXT = 'text'

CLUSTERFUZZ_AUTHORIZATION_HEADER = 'x-clusterfuzz-authorization'
CLUSTERFUZZ_AUTHORIZATION_IDENTITY = 'x-clusterfuzz-identity'
VERIFICATION_CODE_PREFIX = 'VerificationCode '
BEARER_PREFIX = 'Bearer '

_auth_config_obj = None


def _auth_config():
  """Return a config with auth root."""
  global _auth_config_obj
  if not _auth_config_obj:
    _auth_config_obj = local_config.AuthConfig()

  return _auth_config_obj


def extend_request(req, params):
  """Extends a request."""

  def _iterparams():
    for k, v in params.items():
      yield k, v

  def _get(key, default_value=None):
    """Return the value of the key or the default value."""
    return params.get(key, default_value)

  req.get = _get
  req.iterparams = _iterparams


def extend_json_request(req):
  """Extends a request to support JSON."""
  try:
    params = json.loads(req.data)
  except ValueError as e:
    raise helpers.EarlyExitException(
        'Parsing the JSON request body failed: %s' % req.data, 400) from e

  extend_request(req, params)


def cron():
  """Wrap a handler with cron."""

  def decorator(func):
    """Decorator."""

    @functools.wraps(func)
    def wrapper(self):
      """Wrapper."""
      if not self.is_cron():
        raise helpers.AccessDeniedException('You are not a cron.')

      result = func(self)
      if result is None:
        return 'OK'

      return result

    return wrapper

  return decorator


def check_admin_access(func):
  """Wrap a handler with admin checking.

  This decorator must be below post(..) and get(..) when used.
  """

  @functools.wraps(func)
  def wrapper(self):
    """Wrapper."""
    if not auth.is_current_user_admin():
      raise helpers.AccessDeniedException('Admin access is required.')

    return func(self)

  return wrapper


def check_admin_access_if_oss_fuzz(func):
  """Wrap a handler with an admin check if this is OSS-Fuzz.

  This decorator must be below post(..) and get(..) when used.
  """

  @functools.wraps(func)
  def wrapper(self):
    """Wrapper."""
    if utils.is_oss_fuzz():
      return check_admin_access(func)(self)

    return func(self)

  return wrapper


def unsupported_on_local_server(func):
  """Wrap a handler to raise error when running in local App Engine
  development environment.

  This decorator must be below post(..) and get(..) when used.
  """

  @functools.wraps(func)
  def wrapper(self, *args, **kwargs):
    """Wrapper."""
    if environment.is_running_on_app_engine_development():
      raise helpers.EarlyExitException(
          'This feature is not available in local App Engine Development '
          'environment.', 400)

    return func(self, *args, **kwargs)

  return wrapper


def get_access_token(verification_code):
  """Get the access token from verification code.

    See: https://developers.google.com/identity/protocols/OAuth2InstalledApp
  """
  client_id = db_config.get_value('reproduce_tool_client_id')
  if not client_id:
    raise helpers.UnauthorizedException('Client id not configured.')

  client_secret = db_config.get_value('reproduce_tool_client_secret')
  if not client_secret:
    raise helpers.UnauthorizedException('Client secret not configured.')

  response = requests.post(
      'https://www.googleapis.com/oauth2/v4/token',
      headers={'Content-Type': 'application/x-www-form-urlencoded'},
      data={
          'code': verification_code,
          'client_id': client_id,
          'client_secret': client_secret,
          'redirect_uri': 'urn:ietf:wg:oauth:2.0:oob',
          'grant_type': 'authorization_code'
      })

  if response.status_code != 200:
    raise helpers.UnauthorizedException('Invalid verification code (%s): %s' %
                                        (verification_code, response.text))

  try:
    data = json.loads(response.text)
    return data['access_token']
  except (KeyError, ValueError) as e:
    raise helpers.EarlyExitException(
        'Parsing the JSON response body failed: %s' % response.text, 500) from e


def get_email_and_access_token(authorization):
  """Get user email from the request.

    See: https://developers.google.com/identity/protocols/OAuth2InstalledApp
  """
  if authorization.startswith(VERIFICATION_CODE_PREFIX):
    verification_code = authorization.split(' ')[1]
    access_token = get_access_token(verification_code)
    authorization = BEARER_PREFIX + access_token

  if not authorization.startswith(BEARER_PREFIX):
    raise helpers.UnauthorizedException(
        'The Authorization header is invalid. It should have been started with'
        " '%s'." % BEARER_PREFIX)

  access_token = authorization.split(' ')[1]

  response = requests.get(
      'https://www.googleapis.com/oauth2/v3/tokeninfo',
      params={'access_token': access_token})
  if response.status_code != 200:
    raise helpers.UnauthorizedException(
        'Failed to authorize. The Authorization header (%s) might be invalid.' %
        authorization)

  try:
    data = json.loads(response.text)

    # Whitelist service accounts. They have different client IDs (or aud).
    # Therefore, we check against their email directly.
    if data.get('email_verified') and data.get('email') in _auth_config().get(
        'whitelisted_oauth_emails', default=[]):
      return data['email'], authorization

    # Validate that this is an explicitly whitelisted client ID, or the client
    # ID for the reproduce tool.
    whitelisted_client_ids = _auth_config().get(
        'whitelisted_oauth_client_ids', default=[])
    reproduce_tool_client_id = db_config.get_value('reproduce_tool_client_id')
    if reproduce_tool_client_id:
      whitelisted_client_ids += [reproduce_tool_client_id]
    if data.get('aud') not in whitelisted_client_ids:
      raise helpers.UnauthorizedException(
          "The access token doesn't belong to one of the allowed OAuth clients"
          ': %s.' % response.text)

    if not data.get('email_verified'):
      raise helpers.UnauthorizedException('The email (%s) is not verified: %s.'
                                          % (data.get('email'), response.text))

    return data['email'], authorization
  except (KeyError, ValueError) as e:
    raise helpers.EarlyExitException(
        'Parsing the JSON response body failed: %s' % response.text, 500) from e


def oauth(func):
  """Wrap a handler with OAuth authentication by reading the Authorization
    header and getting user email.
  """

  @functools.wraps(func)
  def wrapper(self):
    """Wrapper."""
    auth_header = request.headers.get('Authorization')
    if auth_header:
      email, returned_auth_header = get_email_and_access_token(auth_header)
      setattr(g, '_oauth_email', email)

      response = make_response(func(self))
      response.headers[CLUSTERFUZZ_AUTHORIZATION_HEADER] = str(
          returned_auth_header)
      response.headers[CLUSTERFUZZ_AUTHORIZATION_IDENTITY] = str(email)
      return response

    return func(self)

  return wrapper


def pubsub_push(func):
  """Wrap a handler with pubsub push authentication."""

  @functools.wraps(func)
  def wrapper(self):
    """Wrapper."""
    try:
      bearer_token = request.headers.get('Authorization', '')
      if not bearer_token.startswith(BEARER_PREFIX):
        raise helpers.UnauthorizedException('Missing or invalid bearer token.')

      token = bearer_token.split(' ')[1]
      claim = id_token.verify_oauth2_token(token, google_requests.Request())
    except google.auth.exceptions.GoogleAuthError as e:
      raise helpers.UnauthorizedException('Invalid ID token.') from e

    if (not claim.get('email_verified') or
        claim.get('email') != utils.service_account_email()):
      raise helpers.UnauthorizedException('Invalid ID token.')

    message = pubsub.raw_message_to_message(json.loads(request.data.decode()))
    return func(self, message)

  return wrapper


def check_user_access(need_privileged_access):
  """Wrap a handler with check_user_access.

  This decorator must be below post(..) and get(..) when used.
  """

  def decorator(func):
    """Decorator."""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
      """Wrapper."""
      if not access.has_access(need_privileged_access=need_privileged_access):
        raise helpers.AccessDeniedException()

      return func(self, *args, **kwargs)

    return wrapper

  return decorator


def check_testcase_access(func):
  """Wrap a handler with check_testcase_access.

  It expects the param
    `testcaseId`. And it expects func to have testcase as its first argument.

  This decorator must be below post(..) and get(..) when used.
  """

  @functools.wraps(func)
  def wrapper(self):
    """Wrapper."""
    testcase_id = helpers.cast(
        request.get('testcaseId'), int,
        "The param 'testcaseId' is not a number.")

    testcase = access.check_access_and_get_testcase(testcase_id)
    return func(self, testcase)

  return wrapper


def allowed_cors(func):
  """Wrap a handler with 'Access-Control-Allow-Origin to allow cross-domain
  AJAX calls."""

  @functools.wraps(func)
  def wrapper(self):
    """Wrapper."""
    origin = request.headers.get('Origin')
    whitelisted_cors_urls = _auth_config().get('whitelisted_cors_urls')
    response = make_response(func(self))

    if origin and whitelisted_cors_urls:
      for domain_regex in whitelisted_cors_urls:
        if re.match(domain_regex, origin):
          response.headers['Access-Control-Allow-Origin'] = origin
          response.headers['Vary'] = 'Origin'
          response.headers['Access-Control-Allow-Credentials'] = 'true'
          response.headers['Access-Control-Allow-Methods'] = (
              'GET,OPTIONS,POST')
          response.headers['Access-Control-Allow-Headers'] = (
              'Accept,Authorization,Content-Type')
          response.headers['Access-Control-Max-Age'] = '3600'
          break

    return response

  return wrapper


def post(request_content_type, response_content_type):
  """Wrap a POST handler, parse request, and set response's content type."""

  def decorator(func):
    """Decorator."""

    @functools.wraps(func)
    def wrapper(self):
      """Wrapper."""
      if response_content_type == JSON:
        self.is_json = True

      if request_content_type == JSON:
        extend_json_request(request)
      elif request_content_type == FORM:
        extend_request(request, request.form)
      else:
        extend_request(request, request.args)

      response = make_response(func(self))
      if response_content_type == JSON:
        response.headers['Content-Type'] = 'application/json'
      elif response_content_type == TEXT:
        response.headers['Content-Type'] = 'text/plain'
      elif response_content_type == HTML:
        # Don't enforce content security policies in local development mode.
        if not environment.is_running_on_app_engine_development():
          response.headers['Content-Security-Policy'] = csp.get_default()

      return response

    return wrapper

  return decorator


def get(response_content_type):
  """Wrap a GET handler and set response's content type."""

  def decorator(func):
    """Decorator."""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
      """Wrapper."""
      if response_content_type == JSON:
        self.is_json = True

      extend_request(request, request.args)
      response = make_response(func(self, *args, **kwargs))
      if response_content_type == JSON:
        response.headers['Content-Type'] = 'application/json'
      elif response_content_type == TEXT:
        response.headers['Content-Type'] = 'text/plain'
      elif response_content_type == HTML:
        # Don't enforce content security policies in local development mode.
        if not environment.is_running_on_app_engine_development():
          response.headers['Content-Security-Policy'] = csp.get_default()

      return response

    return wrapper

  return decorator


def require_csrf_token(func):
  """Wrap a handler to require a valid CSRF token."""

  def wrapper(self, *args, **kwargs):
    """Check to see if this handler has a valid CSRF token provided to it."""
    token_value = request.get('csrf_token')
    user = auth.get_current_user()
    if not user:
      raise helpers.AccessDeniedException('Not logged in.')

    query = data_types.CSRFToken.query(
        data_types.CSRFToken.value == token_value,
        data_types.CSRFToken.user_email == user.email)
    token = query.get()
    if not token:
      raise helpers.AccessDeniedException('Invalid CSRF token.')

    # Make sure that the token is not expired.
    if token.expiration_time < datetime.datetime.utcnow():
      token.key.delete()
      raise helpers.AccessDeniedException('Expired CSRF token.')

    return func(self, *args, **kwargs)

  return wrapper



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the stack analyzer module for android specifically."""

import os
import unittest

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_analyzer
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers

DATA_DIRECTORY = os.path.join(os.path.dirname(__file__), 'stack_analyzer_data')
TEST_JOB_NAME = 'test'

KERNEL_REPRO = """kernel/build u'c059b39e1caf2b96aa376582eeb93062b43d69d5'
kernel/manifest u'75a64986ab455f8b45087b8ad54db68bcb8988f4'
kernel/private/msm-google u'40e9b2ff3a280a8775cfcd5841e530ce78f94355'
kernel/private/msm-google-extra/audiokernel u'112a618d5b757b0600c69f7385892b3f57ccd93e'
kernel/private/msm-google-modules/data-kernel u'e7210f09d00c91f87b295c7a952f040c73506cc0'
kernel/private/msm-google-modules/fts_touch u'8f6a4e9f5649deff59174ffed1d5c2af196d9f63'
kernel/private/msm-google-modules/qca-wfi-host-cmn u'7d4b05ac12d6a1b5d5247da35ae7e370a2cba07d'
kernel/private/msm-google-modules/qcacld u'0a077b0073c48555d0edb2b9b0510fb883181828'
kernel/private/msm-google-modules/wlan-fw-api u'53d899727e4278f4e9fb46328d740a8fb2d9a493'
kernel/private/tests/patchwork u'204e78fb6d905016bfc16ebe7b64547f388cfdb5'
kernel/tests u'bfef3bb78b23cb3f3f12a6880ecafd5def3b66a5'
platform/external/fff u'c82edb1fc60dc81bd319d9b8d0bee9f8963a6960'
platform/external/googletest u'a037984aea3317260edd1127abb39e30e845bc94'
platform/prebuilts/clang/host/linux-x86 u'4b1f275e6b3826c86f791ae8c4d5ec3563c2fc11'
platform/prebuilts/gcc/linux-x86/aarch64/aarch64-linux-android-4.9 u'961622e926a1b21382dba4dd9fe0e5fb3ee5ab7c'
platform/prebuilts/gcc/linux-x86/arm/arm-linux-androideabi-4.9 u'cb7b3ac1b7fdb49474ff68761909934d1142f594'
platform/prebuilts/misc u'15560bb32cdb9b47db48eb4865b736df9708a8fe'
platform/tools/repohooks u'233b8010f7f5e3c544b47c68ffae781860156945'
"""


# pylint: disable=unused-argument
def _mock_symbolize_stacktrace(stacktrace, enable_inline_frames=True):
  """No-op mocked version of symbolize_stacktrace."""
  return stacktrace


def _mock_fetch_artifact_get(bid,
                             target,
                             regex,
                             output_directory,
                             output_filename_override=None):
  if output_filename_override:
    artifact_path = os.path.join(output_directory, output_filename_override)
    with open(artifact_path, 'w') as artifact_file:
      artifact_file.write(KERNEL_REPRO)


# pylint: enable=unused-argument


class AndroidStackAnalyzerTest(unittest.TestCase):
  """Android specific Stack analyzer tests."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.crash_analysis.stack_parsing.stack_symbolizer.'
        'symbolize_stacktrace',
        'clusterfuzz._internal.metrics.logs.log_error',
    ])
    os.environ['JOB_NAME'] = TEST_JOB_NAME

    self.mock.symbolize_stacktrace.side_effect = _mock_symbolize_stacktrace

  def _mock_read_data_from_file(self, file_path, eval_data=True, default=None):
    if file_path.endswith('repo.prop'):
      return self._real_read_data_from_file(file_path, eval_data, default)

    return None

  def _read_test_data(self, name):
    """Helper function to read test data."""
    with open(os.path.join(DATA_DIRECTORY, name)) as handle:
      return handle.read()

  def test_syzkaller_kasan_android_with_env(self):
    """Test syzkaller kasan."""
    environment.set_value('OS_OVERRIDE', 'ANDROID_KERNEL')
    environment.set_bot_environment()
    self._real_read_data_from_file = utils.read_data_from_file
    test_helpers.patch(self, [
        'clusterfuzz._internal.platforms.android.fetch_artifact.get',
        'clusterfuzz._internal.platforms.android.kernel_utils.get_kernel_hash_and_build_id',
        'clusterfuzz._internal.platforms.android.kernel_utils.get_kernel_name',
        'clusterfuzz._internal.platforms.android.settings.get_product_brand',
        'clusterfuzz._internal.google_cloud_utils.storage.get_file_from_cache_if_exists',
        'clusterfuzz._internal.google_cloud_utils.storage.store_file_in_cache',
        'clusterfuzz._internal.base.utils.write_data_to_file',
        'clusterfuzz._internal.base.utils.read_data_from_file'
    ])
    self.mock.get.side_effect = _mock_fetch_artifact_get
    self.mock.get_kernel_hash_and_build_id.return_value = '40e9b2ff3a2', '12345'
    self.mock.get_kernel_name.return_value = 'device_kernel'
    self.mock.get_product_brand.return_value = 'google'
    self.mock.get_file_from_cache_if_exists.return_value = False
    self.mock.store_file_in_cache.return_value = None
    self.mock.write_data_to_file = None
    self.mock.read_data_from_file.side_effect = self._mock_read_data_from_file

    data = self._read_test_data('kasan_syzkaller_android.txt')
    expected_stack = self._read_test_data(
        'kasan_syzkaller_android_linkified.txt')
    actual_state = stack_analyzer.get_crash_data(data)
    self.assertEqual(actual_state.crash_stacktrace, expected_stack)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: clusterfuzz/_internal/protos/process_state.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='clusterfuzz/_internal/protos/process_state.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n0clusterfuzz/_internal/protos/process_state.proto\"\x9f\x03\n\x11ProcessStateProto\x12\x17\n\x0ftime_date_stamp\x18\x01 \x01(\x03\x12\x1b\n\x13process_create_time\x18\r \x01(\x03\x12\'\n\x05\x63rash\x18\x02 \x01(\x0b\x32\x18.ProcessStateProto.Crash\x12\x11\n\tassertion\x18\x03 \x01(\t\x12\x19\n\x11requesting_thread\x18\x04 \x01(\x05\x12*\n\x07threads\x18\x05 \x03(\x0b\x32\x19.ProcessStateProto.Thread\x12\x1c\n\x07modules\x18\x06 \x03(\x0b\x32\x0b.CodeModule\x12\n\n\x02os\x18\x07 \x01(\t\x12\x10\n\x08os_short\x18\x08 \x01(\t\x12\x12\n\nos_version\x18\t \x01(\t\x12\x0b\n\x03\x63pu\x18\n \x01(\t\x12\x10\n\x08\x63pu_info\x18\x0b \x01(\t\x12\x11\n\tcpu_count\x18\x0c \x01(\x05\x1a(\n\x05\x43rash\x12\x0e\n\x06reason\x18\x01 \x02(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x02(\x03\x1a%\n\x06Thread\x12\x1b\n\x06\x66rames\x18\x01 \x03(\x0b\x32\x0b.StackFrame\"\x8e\x03\n\nStackFrame\x12\x13\n\x0binstruction\x18\x01 \x02(\x03\x12\x1b\n\x06module\x18\x02 \x01(\x0b\x32\x0b.CodeModule\x12\x15\n\rfunction_name\x18\x03 \x01(\t\x12\x15\n\rfunction_base\x18\x04 \x01(\x03\x12\x18\n\x10source_file_name\x18\x05 \x01(\t\x12\x13\n\x0bsource_line\x18\x06 \x01(\x05\x12\x18\n\x10source_line_base\x18\x07 \x01(\x03\x12%\n\x05trust\x18\x08 \x01(\x0e\x32\x16.StackFrame.FrameTrust\"\xaf\x01\n\nFrameTrust\x12\x14\n\x10\x46RAME_TRUST_NONE\x10\x00\x12\x14\n\x10\x46RAME_TRUST_SCAN\x10\x01\x12\x18\n\x14\x46RAME_TRUST_CFI_SCAN\x10\x02\x12\x12\n\x0e\x46RAME_TRUST_FP\x10\x03\x12\x13\n\x0f\x46RAME_TRUST_CFI\x10\x04\x12\x19\n\x15\x46RAME_TRUST_PREWALKED\x10\x05\x12\x17\n\x13\x46RAME_TRUST_CONTEXT\x10\x06\"\x9b\x01\n\nCodeModule\x12\x14\n\x0c\x62\x61se_address\x18\x01 \x01(\x03\x12\x0c\n\x04size\x18\x02 \x01(\x03\x12\x11\n\tcode_file\x18\x03 \x01(\t\x12\x17\n\x0f\x63ode_identifier\x18\x04 \x01(\t\x12\x12\n\ndebug_file\x18\x05 \x01(\t\x12\x18\n\x10\x64\x65\x62ug_identifier\x18\x06 \x01(\t\x12\x0f\n\x07version\x18\x07 \x01(\t'
)



_STACKFRAME_FRAMETRUST = _descriptor.EnumDescriptor(
  name='FrameTrust',
  full_name='StackFrame.FrameTrust',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FRAME_TRUST_NONE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FRAME_TRUST_SCAN', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FRAME_TRUST_CFI_SCAN', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FRAME_TRUST_FP', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FRAME_TRUST_CFI', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FRAME_TRUST_PREWALKED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FRAME_TRUST_CONTEXT', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=694,
  serialized_end=869,
)
_sym_db.RegisterEnumDescriptor(_STACKFRAME_FRAMETRUST)


_PROCESSSTATEPROTO_CRASH = _descriptor.Descriptor(
  name='Crash',
  full_name='ProcessStateProto.Crash',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='reason', full_name='ProcessStateProto.Crash.reason', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='address', full_name='ProcessStateProto.Crash.address', index=1,
      number=2, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=389,
  serialized_end=429,
)

_PROCESSSTATEPROTO_THREAD = _descriptor.Descriptor(
  name='Thread',
  full_name='ProcessStateProto.Thread',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='frames', full_name='ProcessStateProto.Thread.frames', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=431,
  serialized_end=468,
)

_PROCESSSTATEPROTO = _descriptor.Descriptor(
  name='ProcessStateProto',
  full_name='ProcessStateProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='time_date_stamp', full_name='ProcessStateProto.time_date_stamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='process_create_time', full_name='ProcessStateProto.process_create_time', index=1,
      number=13, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='crash', full_name='ProcessStateProto.crash', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='assertion', full_name='ProcessStateProto.assertion', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='requesting_thread', full_name='ProcessStateProto.requesting_thread', index=4,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='threads', full_name='ProcessStateProto.threads', index=5,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='modules', full_name='ProcessStateProto.modules', index=6,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='os', full_name='ProcessStateProto.os', index=7,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='os_short', full_name='ProcessStateProto.os_short', index=8,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='os_version', full_name='ProcessStateProto.os_version', index=9,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cpu', full_name='ProcessStateProto.cpu', index=10,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cpu_info', full_name='ProcessStateProto.cpu_info', index=11,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cpu_count', full_name='ProcessStateProto.cpu_count', index=12,
      number=12, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PROCESSSTATEPROTO_CRASH, _PROCESSSTATEPROTO_THREAD, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=468,
)


_STACKFRAME = _descriptor.Descriptor(
  name='StackFrame',
  full_name='StackFrame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='instruction', full_name='StackFrame.instruction', index=0,
      number=1, type=3, cpp_type=2, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='module', full_name='StackFrame.module', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='function_name', full_name='StackFrame.function_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='function_base', full_name='StackFrame.function_base', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_file_name', full_name='StackFrame.source_file_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_line', full_name='StackFrame.source_line', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_line_base', full_name='StackFrame.source_line_base', index=6,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trust', full_name='StackFrame.trust', index=7,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _STACKFRAME_FRAMETRUST,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=471,
  serialized_end=869,
)


_CODEMODULE = _descriptor.Descriptor(
  name='CodeModule',
  full_name='CodeModule',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='base_address', full_name='CodeModule.base_address', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='size', full_name='CodeModule.size', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code_file', full_name='CodeModule.code_file', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code_identifier', full_name='CodeModule.code_identifier', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='debug_file', full_name='CodeModule.debug_file', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='debug_identifier', full_name='CodeModule.debug_identifier', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='CodeModule.version', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=872,
  serialized_end=1027,
)

_PROCESSSTATEPROTO_CRASH.containing_type = _PROCESSSTATEPROTO
_PROCESSSTATEPROTO_THREAD.fields_by_name['frames'].message_type = _STACKFRAME
_PROCESSSTATEPROTO_THREAD.containing_type = _PROCESSSTATEPROTO
_PROCESSSTATEPROTO.fields_by_name['crash'].message_type = _PROCESSSTATEPROTO_CRASH
_PROCESSSTATEPROTO.fields_by_name['threads'].message_type = _PROCESSSTATEPROTO_THREAD
_PROCESSSTATEPROTO.fields_by_name['modules'].message_type = _CODEMODULE
_STACKFRAME.fields_by_name['module'].message_type = _CODEMODULE
_STACKFRAME.fields_by_name['trust'].enum_type = _STACKFRAME_FRAMETRUST
_STACKFRAME_FRAMETRUST.containing_type = _STACKFRAME
DESCRIPTOR.message_types_by_name['ProcessStateProto'] = _PROCESSSTATEPROTO
DESCRIPTOR.message_types_by_name['StackFrame'] = _STACKFRAME
DESCRIPTOR.message_types_by_name['CodeModule'] = _CODEMODULE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ProcessStateProto = _reflection.GeneratedProtocolMessageType('ProcessStateProto', (_message.Message,), {

  'Crash' : _reflection.GeneratedProtocolMessageType('Crash', (_message.Message,), {
    'DESCRIPTOR' : _PROCESSSTATEPROTO_CRASH,
    '__module__' : 'clusterfuzz._internal.protos.process_state_pb2'
    # @@protoc_insertion_point(class_scope:ProcessStateProto.Crash)
    })
  ,

  'Thread' : _reflection.GeneratedProtocolMessageType('Thread', (_message.Message,), {
    'DESCRIPTOR' : _PROCESSSTATEPROTO_THREAD,
    '__module__' : 'clusterfuzz._internal.protos.process_state_pb2'
    # @@protoc_insertion_point(class_scope:ProcessStateProto.Thread)
    })
  ,
  'DESCRIPTOR' : _PROCESSSTATEPROTO,
  '__module__' : 'clusterfuzz._internal.protos.process_state_pb2'
  # @@protoc_insertion_point(class_scope:ProcessStateProto)
  })
_sym_db.RegisterMessage(ProcessStateProto)
_sym_db.RegisterMessage(ProcessStateProto.Crash)
_sym_db.RegisterMessage(ProcessStateProto.Thread)

StackFrame = _reflection.GeneratedProtocolMessageType('StackFrame', (_message.Message,), {
  'DESCRIPTOR' : _STACKFRAME,
  '__module__' : 'clusterfuzz._internal.protos.process_state_pb2'
  # @@protoc_insertion_point(class_scope:StackFrame)
  })
_sym_db.RegisterMessage(StackFrame)

CodeModule = _reflection.GeneratedProtocolMessageType('CodeModule', (_message.Message,), {
  'DESCRIPTOR' : _CODEMODULE,
  '__module__' : 'clusterfuzz._internal.protos.process_state_pb2'
  # @@protoc_insertion_point(class_scope:CodeModule)
  })
_sym_db.RegisterMessage(CodeModule)


# @@protoc_insertion_point(module_scope)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""utils tests."""
import datetime
import os
import shutil
import sys
import tempfile
import time
import unittest

from google.cloud import ndb
import mock
from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class GetSizeStringTest(unittest.TestCase):
  """Test get_size_string."""

  def test_size_bytes(self):
    """Test size returned in bytes."""
    self.assertEqual(utils.get_size_string(10), '10 B')
    self.assertEqual(utils.get_size_string(1023), '1023 B')

  def test_size_kilobytes(self):
    """Test size returned in kilobytes."""
    self.assertEqual(utils.get_size_string(1024), '1 KB')
    self.assertEqual(utils.get_size_string(1048575), '1023 KB')

  def test_size_megabytes(self):
    """Test size returned in megabytes."""
    self.assertEqual(utils.get_size_string(1048576), '1 MB')
    self.assertEqual(utils.get_size_string(1073741823), '1023 MB')

  def test_size_gigabytes(self):
    """Test size returned in gigabytes."""
    self.assertEqual(utils.get_size_string(1073741824), '1 GB')
    self.assertEqual(utils.get_size_string(5000000000), '4 GB')


class GetLineCountString(unittest.TestCase):
  """Test get_line_count_string."""

  def test_line_count_0(self):
    """Test line count string for 0 lines."""
    self.assertEqual(utils.get_line_count_string(0), 'empty')

  def test_line_count_1(self):
    """Test line count string for 1 line."""
    self.assertEqual(utils.get_line_count_string(1), '1 line')

  def test_line_count_10(self):
    """Test line count string for more than 1 line."""
    self.assertEqual(utils.get_line_count_string(10), '10 lines')


@test_utils.with_cloud_emulators('datastore')
class HashTest(unittest.TestCase):
  """Tests for hash helper functions."""

  class DummyEntity(ndb.Model):
    string_property = ndb.StringProperty()
    datetime_property = ndb.DateTimeProperty()
    integer_property = ndb.IntegerProperty()

  def setUp(self):
    entity_1 = HashTest.DummyEntity()
    entity_1.string_proprety = 'abc'
    entity_1.datetime_property = datetime.datetime(2018, 1, 1)
    entity_1.integer_property = 1
    entity_1.put()

    entity_2 = HashTest.DummyEntity()
    entity_2.string_proprety = 'abcd'
    entity_2.datetime_property = datetime.datetime(2018, 1, 1)
    entity_2.integer_property = 2
    entity_2.put()

  def test_hash_equal_for_identical_entities(self):
    """Ensure that two equivalent entities have the same hash value."""
    entity_1 = HashTest.DummyEntity.query(
        HashTest.DummyEntity.integer_property == 1).get()
    hash_1 = utils.entity_hash(entity_1)

    entity_2 = HashTest.DummyEntity.query(
        HashTest.DummyEntity.integer_property == 1).get()
    hash_2 = utils.entity_hash(entity_2)

    self.assertEqual(hash_1, hash_2)

  def test_hash_different_for_different_entities(self):
    """Ensure that two different entities have different hash values."""
    entity_1 = HashTest.DummyEntity.query(
        HashTest.DummyEntity.integer_property == 1).get()
    hash_1 = utils.entity_hash(entity_1)

    entity_2 = HashTest.DummyEntity.query(
        HashTest.DummyEntity.integer_property == 2).get()
    hash_2 = utils.entity_hash(entity_2)

    self.assertNotEqual(hash_1, hash_2)


@test_utils.integration
class TimeoutTest(unittest.TestCase):
  """Test timeout decorator."""

  def setUp(self):
    self.real_time_sleep = time.sleep

    patcher = mock.patch('time.sleep', autospec=True)
    self.time_sleep = patcher.start()
    self.addCleanup(patcher.stop)

  @utils.timeout(1)
  def _func(self, duration):
    self.real_time_sleep(duration)

  def test_exceed_timeout(self):
    """Test function that executes longer than the timeout."""
    self._func(0)

    with self.assertRaises(SystemExit):
      self._func(5)

  @utils.timeout(5)
  @utils.timeout(5)
  def _nested_func(self):
    pass

  def test_nested_does_not_deadlock(self):
    """Test nesting the timeout decorator doesn't cause a deadlock."""
    self._nested_func()


class FilePathToFileUrlTest(unittest.TestCase):
  """Tests file_path_to_file_url."""

  def test_empty(self):
    """Tests empty path."""
    self.assertEqual('', utils.file_path_to_file_url(''))

  def test_window_prefix(self):
    """Tests path with windows prefix."""
    self.assertIn(
        utils.file_path_to_file_url('\\\\?\\c:\\test\\test2.html'),
        [
            'file:///c:/test/test2.html',  # On windows.
            'file:///c%3A%5Ctest%5Ctest2.html',  # On Posix.
        ])


class NormalizePathTest(unittest.TestCase):
  """Tests normalize_path."""

  def test_redundant_path(self):
    """Tests redundant paths."""
    self.assertEqual('/test/test2/test3',
                     utils.normalize_path('/test/./test2/test3'))
    self.assertEqual('/test/test2/test3',
                     utils.normalize_path('/test//test2/test3'))

  def test_normalize_case(self):
    """Tests normalize case for windows."""
    if sys.platform.startswith('win'):
      self.assertEqual(
          utils.normalize_path('C:\\test\\test2\\test3'),
          # Notice that the path starts with a lowercase c.
          utils.normalize_path('c:\\test\\test2\\test3'))
    else:
      self.assertEqual('/mnt/Test', utils.normalize_path('/mnt/Test'))


class NormalizeEmailTest(unittest.TestCase):
  """Tests email normalization."""

  def test_normalize_email(self):
    """Test normalize email."""
    self.assertEqual('a@b.com', utils.normalize_email('a@b.com'))
    self.assertEqual('a@b.com', utils.normalize_email('A@B.com'))

  def test_email_equals(self):
    """Test email comparison."""
    self.assertTrue(utils.emails_equal('a@b.com', 'a@b.com'))
    self.assertTrue(utils.emails_equal('A@b.com', 'a@B.com'))


class RandomWeightedChoiceTest(unittest.TestCase):
  """Tests random_weighted_choice."""

  def setUp(self):
    helpers.patch(self, ['random.SystemRandom.uniform'])

    class O(object):

      def __init__(self, data, weight):
        self.data = data
        self.weight = weight

    self.list = [O('A1', 0), O('A2', 1), O('A3', 3), O('A4', 0), O('A5', 5)]

  def test_1(self):
    self.mock.uniform.return_value = 0
    self.assertNotEqual('A1', utils.random_weighted_choice(self.list).data)

  def test_2(self):
    self.mock.uniform.return_value = 0
    self.assertEqual('A2', utils.random_weighted_choice(self.list).data)

    self.mock.uniform.return_value = 0.4
    self.assertEqual('A2', utils.random_weighted_choice(self.list).data)

    self.mock.uniform.return_value = 1
    self.assertEqual('A2', utils.random_weighted_choice(self.list).data)

  def test_3(self):
    self.mock.uniform.return_value = 1.1
    self.assertEqual('A3', utils.random_weighted_choice(self.list).data)

    self.mock.uniform.return_value = 2
    self.assertEqual('A3', utils.random_weighted_choice(self.list).data)

    self.mock.uniform.return_value = 4
    self.assertEqual('A3', utils.random_weighted_choice(self.list).data)

  def test_4(self):
    self.mock.uniform.return_value = 4.1
    self.assertNotEqual('A4', utils.random_weighted_choice(self.list).data)

  def test_5(self):
    self.mock.uniform.return_value = 4.1
    self.assertEqual('A5', utils.random_weighted_choice(self.list).data)

    self.mock.uniform.return_value = 7
    self.assertEqual('A5', utils.random_weighted_choice(self.list).data)

    self.mock.uniform.return_value = 9
    self.assertEqual('A5', utils.random_weighted_choice(self.list).data)


class GetCrashStacktraceOutputTest(unittest.TestCase):
  """Tests get_crash_stacktrace_output."""

  def setUp(self):
    helpers.patch_environ(self)

    os.environ['TOOL_NAME'] = 'ASAN'
    os.environ['JOB_NAME'] = 'linux_asan_chrome'

    self.start_separator = '+' + '-' * 40
    self.end_separator = '-' * 40 + '+'

  def test_env_settings(self):
    """Tests that environment settings are added."""
    os.environ['ASAN_OPTIONS'] = 'setting1=value1:setting2=value_2'
    self.assertEqual(
        '[Environment] ASAN_OPTIONS=setting1=value1:setting2=value_2\n'
        '[Command line] cmd_line\n\n' + self.start_separator +
        'Release Build Stacktrace' + self.end_separator + '\nsym_stack',
        utils.get_crash_stacktrace_output('cmd_line', 'sym_stack'))

  def test_release_sym_stack(self):
    """Tests release build with a symbolized stack, with build type not
    explicitly passed."""
    self.assertEqual(
        '[Command line] cmd_line_release\n\n' + self.start_separator +
        'Release Build Stacktrace' + self.end_separator + '\nsym_stack',
        utils.get_crash_stacktrace_output('cmd_line_release', 'sym_stack'))

  def test_debug_sym_stack(self):
    """Tests debug build with a symbolized stack, with build type explicitly
    passed."""
    self.assertEqual(
        '[Command line] cmd_line\n\n' + self.start_separator +
        'Debug Build Stacktrace' + self.end_separator + '\nsym_stack',
        utils.get_crash_stacktrace_output(
            'cmd_line', 'sym_stack', build_type='debug'))

  def test_debug_sym_stack_2(self):
    """Tests debug build with a symbolized stack, with build type not explicitly
    passed."""
    self.assertEqual(
        '[Command line] cmd_line_dbg\n\n' + self.start_separator +
        'Debug Build Stacktrace' + self.end_separator + '\nsym_stack',
        utils.get_crash_stacktrace_output('cmd_line_dbg', 'sym_stack'))

  def test_stable_sym_stack(self):
    """Tests stable build with a symbolized stack, with build type not
    explicitly passed."""
    self.assertEqual(
        '[Command line] cmd_line_stable\n\n' + self.start_separator +
        'Stable Build Stacktrace' + self.end_separator + '\nsym_stack',
        utils.get_crash_stacktrace_output('cmd_line_stable', 'sym_stack'))

  def test_beta_sym_stack(self):
    """Tests beta build with a symbolized stack, with build type not
    explicitly passed."""
    self.assertEqual(
        '[Command line] cmd_line_beta\n\n' + self.start_separator +
        'Beta Build Stacktrace' + self.end_separator + '\nsym_stack',
        utils.get_crash_stacktrace_output('cmd_line_beta', 'sym_stack'))

  def test_release_sym_and_unsym_stacks(self):
    """Tests release build with symbolized and unsymbolized stacks, with build
    type not explicitly passed."""
    self.assertEqual(
        '[Command line] cmd_line\n\n' + self.start_separator +
        'Release Build Stacktrace' + self.end_separator + '\nsym_stack\n\n' +
        self.start_separator + 'Release Build Unsymbolized Stacktrace (diff)' +
        self.end_separator + '\n\nunsym_stack',
        utils.get_crash_stacktrace_output('cmd_line', 'sym_stack',
                                          'unsym_stack'))

  def test_debug_sym_and_unsym_stacks(self):
    """Tests debug build with symbolized and unsymbolized stacks, with build
    type explicitly passed."""
    self.assertEqual(
        '[Command line] cmd_line\n\n' + self.start_separator +
        'Debug Build Stacktrace' + self.end_separator + '\nsym_stack\n\n' +
        self.start_separator + 'Debug Build Unsymbolized Stacktrace (diff)' +
        self.end_separator + '\n\nunsym_stack',
        utils.get_crash_stacktrace_output(
            'cmd_line', 'sym_stack', 'unsym_stack', build_type='debug'))

  def test_release_sym_and_unsym_diff_stacks(self):
    """Tests release build with symbolized and unsymbolized stacks, having some
    common frames, and build type not explicitly passed."""
    self.assertEqual(
        '[Command line] cmd_line\n\n' + self.start_separator +
        'Release Build Stacktrace' + self.end_separator +
        '\nc1\nc2\nc3\nsym_stack\nc4\nc5\nc6\n\n' + self.start_separator +
        'Release Build Unsymbolized Stacktrace (diff)' + self.end_separator +
        '\n\nc2\nc3\nunsym_stack\nc4\nc5',
        utils.get_crash_stacktrace_output(
            'cmd_line', 'c1\nc2\nc3\nsym_stack\nc4\nc5\nc6',
            'c1\nc2\nc3\nunsym_stack\nc4\nc5\nc6'))

  def test_debug_sym_and_unsym_diff_stacks(self):
    """Tests debug build with symbolized and unsymbolized stacks, having some
    common frames, and build type explicitly passed."""
    self.assertEqual(
        '[Command line] cmd_line\n\n' + self.start_separator +
        'Debug Build Stacktrace' + self.end_separator +
        '\nc1\nc2\nc3\nsym_stack\nc4\nc5\nc6\n\n' + self.start_separator +
        'Debug Build Unsymbolized Stacktrace (diff)' + self.end_separator +
        '\n\nc2\nc3\nunsym_stack\nc4\nc5',
        utils.get_crash_stacktrace_output(
            'cmd_line',
            'c1\nc2\nc3\nsym_stack\nc4\nc5\nc6',
            'c1\nc2\nc3\nunsym_stack\nc4\nc5\nc6',
            build_type='debug'))


class GetApplicationIDTest(unittest.TestCase):
  """Tests get_application_id."""

  def setUp(self):
    helpers.patch_environ(self)

  def test_no_app_id(self):
    """Test with no app id set in environment."""
    del os.environ['APPLICATION_ID']
    self.assertEqual(None, utils.get_application_id())

  def test_simple_app_id(self):
    """Test simple app id without domain or partition separator."""
    os.environ['APPLICATION_ID'] = 'app_id'
    self.assertEqual('app_id', utils.get_application_id())

  def test_app_id_with_partition(self):
    """Test app id with partition separator, but no domain separator."""
    os.environ['APPLICATION_ID'] = 'dev~app_id'
    self.assertEqual('app_id', utils.get_application_id())

  def test_app_id_with_domain(self):
    """Test app id with domain separator, but no partition separator."""
    os.environ['APPLICATION_ID'] = 'company:app_id'
    self.assertEqual('company:app_id', utils.get_application_id())

  def test_app_id_with_domain_and_partition(self):
    """Test app id with domain and partition separator."""
    os.environ['APPLICATION_ID'] = 's~company:app_id'
    self.assertEqual('company:app_id', utils.get_application_id())


class DefaultProjectNameTest(unittest.TestCase):
  """Tests default_project_name."""

  def setUp(self):
    helpers.patch_environ(self)

  def test_default(self):
    """Test that it returns default project with no environment changes."""
    self.assertEqual('test-project', utils.default_project_name())

  def test_overridden(self):
    """Test that env variable PROJECT_NAME does not affect default project
    name and we still default project."""
    os.environ['PROJECT_NAME'] = 'other-project'
    self.assertEqual('test-project', utils.default_project_name())


class CurrentProjectTest(unittest.TestCase):
  """Tests current_project."""

  def setUp(self):
    helpers.patch_environ(self)

  def test_default(self):
    """Test that it returns default project with no environment changes."""
    self.assertEqual('test-project', utils.current_project())

  def test_overridden(self):
    """Test that env variable PROJECT_NAME is used for current project."""
    os.environ['PROJECT_NAME'] = 'other-project'
    self.assertEqual('other-project', utils.current_project())


class FileHashTest(fake_filesystem_unittest.TestCase):
  """Test file_hash."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    self.test_file = '/test'

  def test_empty_string(self):
    with open(self.test_file, 'wb'):
      pass
    self.assertEqual('da39a3ee5e6b4b0d3255bfef95601890afd80709',
                     utils.file_hash(self.test_file))

  def test_shorter_than_one_chunk(self):
    with open(self.test_file, 'wb') as file_handle:
      file_handle.write(b'ABC')
    self.assertEqual('3c01bdbb26f358bab27f267924aa2c9a03fcfdb8',
                     utils.file_hash(self.test_file))

  def test_longer_than_one_chunk(self):
    with open(self.test_file, 'wb') as file_handle:
      file_handle.write(b'A' * 60000)
    self.assertEqual('8360c01cef8aa7001d1dd8964b9921d4c187da29',
                     utils.file_hash(self.test_file))


class ServiceAccountEmailTest(unittest.TestCase):
  """Tests service_account_email"""

  def setUp(self):
    helpers.patch_environ(self)

  def test_plain_project_id(self):
    """Test with a plain project ID."""
    os.environ['APPLICATION_ID'] = 'project-id'
    self.assertEqual('project-id@appspot.gserviceaccount.com',
                     utils.service_account_email())

  def test_with_domain(self):
    """Test with a project ID with a domain."""
    os.environ['APPLICATION_ID'] = 'domain.com:project-id'
    self.assertEqual('project-id.domain.com@appspot.gserviceaccount.com',
                     utils.service_account_email())


class SearchBytesInFileTest(unittest.TestCase):
  """Tests search_bytes_in_file."""

  def setUp(self):
    self.temp_dir = tempfile.mkdtemp()
    self.test_path = os.path.join(self.temp_dir, 'file')
    with open(self.test_path, 'wb') as f:
      f.write(b'A' * 16 + b'B' * 16 + b'C' + b'D' * 16)

  def tearDown(self):
    shutil.rmtree(self.temp_dir, ignore_errors=True)

  def test_exists(self):
    """Test exists."""
    with open(self.test_path, 'rb') as f:
      self.assertTrue(utils.search_bytes_in_file(b'A', f))

    with open(self.test_path, 'rb') as f:
      self.assertTrue(utils.search_bytes_in_file(b'B', f))

    with open(self.test_path, 'rb') as f:
      self.assertTrue(utils.search_bytes_in_file(b'C', f))

    with open(self.test_path, 'rb') as f:
      self.assertTrue(utils.search_bytes_in_file(b'D', f))

    with open(self.test_path, 'rb') as f:
      self.assertTrue(utils.search_bytes_in_file(b'A' * 16, f))

    with open(self.test_path, 'rb') as f:
      self.assertTrue(utils.search_bytes_in_file(b'B' * 16, f))

    with open(self.test_path, 'rb') as f:
      self.assertTrue(utils.search_bytes_in_file(b'D' * 16, f))

  def test_not_exists(self):
    """Test not exists."""
    with open(self.test_path, 'rb') as f:
      self.assertFalse(utils.search_bytes_in_file(b'A' * 17, f))

    with open(self.test_path, 'rb') as f:
      self.assertFalse(utils.search_bytes_in_file(b'B' * 17, f))

    with open(self.test_path, 'rb') as f:
      self.assertFalse(utils.search_bytes_in_file(b'C' * 2, f))

    with open(self.test_path, 'rb') as f:
      self.assertFalse(utils.search_bytes_in_file(b'D' * 17, f))

    with open(self.test_path, 'rb') as f:
      self.assertFalse(utils.search_bytes_in_file(b'ABCD', f))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Settings change related functions."""
import re

from clusterfuzz._internal.system import environment

from . import adb

BUILD_FINGERPRINT_REGEX = re.compile(
    r'(?P<vendor>.+)\/(?P<target>.+)'
    r'\/(?P<flavor>.+)\/(?P<name_name>.+)'
    r'\/(?P<build_id>.+):(?P<type>.+)\/(?P<keys>.+)')


def change_se_linux_to_permissive_mode():
  """Switch SELinux to permissive mode for working around local file access and
  other issues."""
  adb.run_shell_command(['setenforce', '0'])


def get_build_fingerprint():
  """Return build's fingerprint."""
  return adb.get_property('ro.build.fingerprint')


def get_build_flavor():
  """Return the build flavor."""
  return adb.get_property('ro.build.flavor')


def get_build_parameters():
  """Return build_id, target and type from the device's fingerprint"""
  build_fingerprint = environment.get_value('BUILD_FINGERPRINT',
                                            get_build_fingerprint())
  build_fingerprint_match = BUILD_FINGERPRINT_REGEX.match(build_fingerprint)
  if not build_fingerprint_match:
    return None

  build_id = build_fingerprint_match.group('build_id')
  target = build_fingerprint_match.group('target')
  build_type = build_fingerprint_match.group('type')
  return {'build_id': build_id, 'target': target, 'type': build_type}


def get_build_version():
  """Return the build version of the system as a character.
  K = Kitkat, L = Lollipop, M = Marshmellow, MASTER = Master.
  """
  build_version = adb.get_property('ro.build.id')
  if not build_version:
    return None

  if build_version == 'MASTER':
    return build_version

  match = re.match('^([A-Z])', build_version)
  if not match:
    return None

  return match.group(1)


def get_cpu_arch():
  """Return cpu architecture."""
  return adb.get_property('ro.product.cpu.abi')


def get_device_codename():
  """Return the device codename."""
  serial = environment.get_value('ANDROID_SERIAL')
  devices_output = adb.run_command(['devices', '-l'])

  serial_pattern = r'(^|\s){serial}\s'.format(serial=re.escape(serial))
  serial_regex = re.compile(serial_pattern)

  for line in devices_output.splitlines():
    values = line.strip().split()

    if not serial_regex.search(line):
      continue

    for value in values:
      if not value.startswith('device:'):
        continue
      device_codename = value.split(':')[-1]
      if device_codename:
        return device_codename

  # Unable to get code name.
  return ''


def get_platform_id():
  """Return a string as |android:{codename}_{sanitizer}:{build_version}|."""
  platform_id = 'android'

  # Add codename and sanitizer tool information.
  platform_id += ':%s' % get_device_codename()
  sanitizer_tool_name = get_sanitizer_tool_name()
  if sanitizer_tool_name:
    platform_id += '_%s' % sanitizer_tool_name

  # Add build version.
  build_version = get_build_version()
  if build_version:
    platform_id += ':%s' % build_version

  return platform_id


def get_product_brand():
  """Return product's brand."""
  return adb.get_property('ro.product.brand')


def get_product_name():
  """Return product's name."""
  return adb.get_property('ro.product.name')


def get_product_model():
  """Return product's model."""
  return adb.get_property('ro.product.model')


def get_build_product():
  """Return builds's product."""
  return adb.get_property('ro.build.product')


def get_sanitizer_tool_name():
  """Return sanitizer tool name e.g. ASAN if found on device."""
  build_flavor = get_build_flavor()
  if 'hwasan' in build_flavor:
    return 'hwasan'
  if 'kasan' in build_flavor:
    return 'kasan'
  if 'asan' in build_flavor:
    return 'asan'

  return None


def get_security_patch_level():
  """Return the security patch level reported by the device."""
  return adb.get_property('ro.build.version.security_patch')


def get_kernel_version_string():
  return adb.run_shell_command('cat /proc/version').strip()


def is_google_device():
  """Return true if this is a google branded device."""
  # If a build branch is already set, then this is a Google device. No need to
  # query device which can fail if the device is failing on recovery mode.
  build_branch = environment.get_value('BUILD_BRANCH')
  if build_branch:
    return True

  product_brand = environment.get_value('PRODUCT_BRAND', get_product_brand())
  if product_brand is None:
    return None

  return product_brand in ('google', 'generic')


def is_automotive():
  """Returns if we are running in Android Automotive OS, currently only for
  Osprey or Seahawk."""
  product_model = get_product_model()
  return product_model in ('Osprey', 'Seahawk')


def set_content_setting(table, key, value):
  """Set a device content setting. The input is not sanitized, so make sure to
  use with trusted input key and value pair only."""

  def _get_type_binding(value):
    """Return binding type for content setting."""
    if isinstance(value, bool):
      return 'b'
    if isinstance(value, float):
      return 'f'
    if isinstance(value, int):
      return 'i'

    # Default to string.
    return 's'

  content_setting_command = (
      'content insert --uri content://%s --bind name:s:%s --bind value:%s:%s' %
      (table, key, _get_type_binding(value), str(value)))

  adb.run_shell_command(content_setting_command)


def set_database_setting(database_path, table, key, value):
  """Update a key in a database. The input is not sanitized, so make sure to use
  with trusted input key and value pair only."""
  sql_command_string = ('"UPDATE %s SET value=\'%s\' WHERE name=\'%s\'"') % (
      table, str(value), key)
  adb.run_shell_command(['sqlite3', database_path, sql_command_string])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Translation from Java code for JavaScriptBaseLexer made to work with
JavaScriptLexer"""

import antlr4


class JavaScriptBaseLexer(antlr4.Lexer):

  def __init__(self, input, output):
    super(JavaScriptBaseLexer, self).__init__(input, output)
    self._scopeStrictModes = []
    self._lastToken = None
    self._useStrictDefault = False
    self._useStrictCurrent = False

  def IsStartOfFile(self):
    return self._lastToken == None

  def getStrictDefault(self):
    return self._useStrictDefault

  def setUseStrictDefault(self, bool):
    self._useStrictDefault = bool
    self._useStrictCurrent = bool

  def IsStrictMode(self):
    return self._useStrictCurrent

  def nextToken(self):
    next = super(JavaScriptBaseLexer, self).nextToken()
    if (next.channel == antlr4.Token.DEFAULT_CHANNEL):
      self._lastToken = next

    return next

  def ProcessOpenBrace(self):
    if len(self._scopeStrictModes) > 0 and self._scopeStrictModes[-1]:
      self._useStrictCurrent = True
    else:
      self._useStrictCurrent = self._useStrictDefault

    self._scopeStrictModes.append(self._useStrictCurrent)

  def ProcessCloseBrace(self):
    if len(self._scopeStrictModes) > 0:
      self._useStrictCurrent = self._scopeStrictModes.pop()
    else:
      self._useStrictCurrent = self._useStrictDefault

  def ProcessStringLiteral(self):
    if self._lastToken == None or self._lastToken.type == self.OpenBrace:
      text = super(JavaScriptBaseLexer, self).text
      if text == '"use strict"' or text == '\'use strict\'':
        if len(self._scopeStrictModes) > 0:
          self._scopeStrictModes.pop()
        self._useStrictCurrent = True
        self._scopeStrictModes.append(self._useStrictCurrent)

  def IsRegExPossible(self):
    if self._lastToken == None:
      return True

    if self._lastToken in [
        self.Identifier, self.NullLiteral, self.BooleanLiteral, self.This,
        self.CloseBracket, self.CloseParen, self.OctalIntegerLiteral,
        self.StringLiteral, self.PlusPlus, self.MinusMinus
    ]:
      return False
    return True



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for jobs."""
import unittest

import flask
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers import crash_query

TEST_STACKTRACE_OVERFLOW = '''
=================================================================
==14479==ERROR: AddressSanitizer: heap-buffer-overflow on address 0x615000000c80 at pc 0x000000c43002 bp 0x7ffc7ba6fae0 sp 0x7ffc7ba6fad8
WRITE of size 4 at 0x615000000c80 thread T0
SCARINESS: 36 (4-byte-write-heap-buffer-overflow)
    #0 0xc48001 in Foo foo.c:36:4
    #1 0x3d912f in Bar bar.c:2:6
    #2 0x41e77e in Main main.c:47:13
    #3 0x3f0778 in LLVMFuzzerTestOneInput fuzzer.cpp:12:1
    #4 0x4f216b in main
    #5 0x7f371939782f in __libc_start_main /build/glibc-LK5gWL/glibc-2.23/csu/libc-start.c:291
    #6 0x441bb8 in _start
SUMMARY: AddressSanitizer: heap-buffer-overflow ()
==14479==ABORTING
'''

TEST_STACKTRACE_OOM = '''
==1== ERROR: libFuzzer: out-of-memory (used: 2590Mb; limit: 2560Mb)
   To change the out-of-memory limit use -rss_limit_mb=<N>
SUMMARY: libFuzzer: out-of-memory
'''


@test_utils.with_cloud_emulators('datastore')
class CrashQueryTest(unittest.TestCase):
  """Jobs tests."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.auth.get_current_user',
    ])
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=crash_query.Handler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

  def test_new(self):
    """Test new."""
    response = self.app.post_json(
        '/', {
            'project': 'project',
            'fuzz_target': 'target',
            'stacktrace': TEST_STACKTRACE_OVERFLOW,
        })

    self.assertEqual({
        'result': 'new',
        'type': 'Heap-buffer-overflow\nWRITE 4',
        'state': 'Foo\nBar\nMain\n',
        'security': True,
    }, response.json)

  def test_duplicate(self):
    """Test duplicate."""
    expected_crash_state = 'Foo\nBar\nMain\n'
    t = data_types.Testcase(
        open=True,
        status='Processed',
        crash_state=expected_crash_state,
        crash_type='Heap-buffer-overflow\nWRITE 4',
        project_name='project',
        security_flag=True)
    t.put()

    response = self.app.post_json(
        '/', {
            'project': 'project',
            'fuzz_target': 'target',
            'stacktrace': TEST_STACKTRACE_OVERFLOW,
        })
    self.assertEqual({
        'result': 'duplicate',
        'duplicate_id': 1,
        'type': 'Heap-buffer-overflow\nWRITE 4',
        'state': expected_crash_state,
        'security': True,
    }, response.json)

    t.group_bug_information = 123
    t.put()
    response = self.app.post_json(
        '/', {
            'project': 'project',
            'fuzz_target': 'target',
            'stacktrace': TEST_STACKTRACE_OVERFLOW,
        })
    self.assertEqual({
        'result': 'duplicate',
        'duplicate_id': 1,
        'bug_id': '123',
        'type': 'Heap-buffer-overflow\nWRITE 4',
        'state': expected_crash_state,
        'security': True,
    }, response.json)

    t.bug_information = '1337'
    t.put()
    response = self.app.post_json(
        '/', {
            'project': 'project',
            'fuzz_target': 'target',
            'stacktrace': TEST_STACKTRACE_OVERFLOW,
        })
    self.assertEqual({
        'result': 'duplicate',
        'duplicate_id': 1,
        'bug_id': '1337',
        'type': 'Heap-buffer-overflow\nWRITE 4',
        'state': expected_crash_state,
        'security': True,
    }, response.json)

  def test_oom(self):
    """Test OOM parsing."""
    response = self.app.post_json(
        '/', {
            'project': 'project',
            'fuzz_target': 'target',
            'stacktrace': TEST_STACKTRACE_OOM,
        })
    self.assertEqual({
        'result': 'new',
        'type': 'Out-of-memory',
        'state': 'target\n',
        'security': False,
    }, response.json)
